// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';
import 'dart:typed_data';

import '../../src/document/document.dart';
import '../../src/store/index_directory.dart';

/// IndexReader is providing an interface for accessing a point-in-time view of an index.
///
/// Any changes made to the index via IndexWriter will not be visible until a new IndexReader is opened.
/// It's best to use [DirectoryReader.open(IndexDirectory)] to obtain an IndexReader.
class IndexReader {
  LiveDocsSource _liveDocsSource;
  IndexSource _indexSource;
  DataSource _dataSource;

  DocSet thisLiveDocs = new DocSet(new Set<int>());
  Map<String, Map<String, DocSet>> thisIndexes =
      <String, Map<String, DocSet>>{};
  Map<String, Map<int, DocData>> thisData = <String, Map<int, DocData>>{};

  IndexReader(IndexDirectory directory) {
    directory.initInput();
    _liveDocsSource = directory.inputLiveDocs();
    _indexSource = directory.inputIndex();
    _dataSource = directory.inputData();
  }

  Future<Null> read() async {
    if (thisLiveDocs.length == 0) {
      thisLiveDocs = await _liveDocsSource.read(thisLiveDocs);
    }
    if (thisIndexes.length == 0) {
      thisIndexes = await _indexSource.read(thisIndexes);
    }
    if (thisData.length == 0) {
      thisData = await _dataSource.read(thisData);
    }
  }

  Future<DocSet> liveDocs() async {
    if (thisLiveDocs.length == 0) {
      thisLiveDocs = await _liveDocsSource.read(thisLiveDocs);
    }
    return thisLiveDocs;
  }

  Future<Iterable<String>> terms(String field) async {
    if (thisIndexes.length == 0 || thisIndexes[field] == null) {
      thisIndexes = await _indexSource.terms(field, thisIndexes);
    }
    final Map<String, DocSet> r = thisIndexes[field];
    return (r == null) ? <String>[] : r.keys;
  }

  Future<DocSet> postings(String field, String term) async {
    if (thisIndexes.length == 0 ||
        thisIndexes[field] == null ||
        thisIndexes[field][term] == null) {
      thisIndexes = await _indexSource.postings(field, term, thisIndexes);
    }
    final DocSet r = thisIndexes[field][term];
    return (r == null) ? new DocSet(new Set<int>()) : r;
  }

  Future<Iterable<String>> dataFields() async {
    if (thisData.length == 0) {
      thisData = await _dataSource.fields(thisData);
    }
    final Iterable<String> r = thisData.keys;
    return (r == null) ? <String>[] : r;
  }

  Future<List<String>> dataValues(String field, int docId) async {
    if (thisData.length == 0 ||
        thisData[field] == null ||
        thisData[field][docId] == null) {
      thisData = await _dataSource.values(field, docId, thisData);
    }
    final DocData r = thisData[field][docId];
    return (r == null) ? <String>[] : r.toList();
  }

  Future<int> docFreq(String field, String text) async {
    final DocSet docSet = await postings(field, text);
    final Set<int> r = docSet.toSet();
    return (r == null) ? -1 : r.length;
  }

  Future<int> maxDoc() async {
    final List<String> r = (await terms('id')).toList();
    return (r == null) ? 0 : r.length;
  }

  Future<Document> document(int docId, {Iterable<String> fieldsToLoad}) async {
    final Iterable<String> fields =
        (fieldsToLoad != null) ? fieldsToLoad : await dataFields();
    final Document ret = new Document();
    for (String field in fields) {
      final List<String> text = await dataValues(field, docId);
      ret.add(new Field(field, text, null, null));
    }
    return ret;
  }
}

class DirectoryReader {
  static Future<IndexReader> open(IndexDirectory directory) async {
    final IndexReader ir = new IndexReader(directory);
    await ir.read();
    return ir;
  }
}

class DocSet {
  Uint32List buffer;
  int length;
  DocSet(Set<int> docIds) {
    buffer = new Uint32List.fromList(docIds.toList());
    length = docIds.length;
  }
  void update(Set<int> docIds) {
    buffer = new Uint32List.fromList(docIds.toList());
    length = docIds.length;
  }

  Set<int> toSet() => buffer.toSet();
}

class DocData {
  Uint16List buffer;
  DocData(List<String> data) {
    buffer = new Uint16List.fromList(data.join('\$').codeUnits);
  }
  List<String> toList() => new String.fromCharCodes(buffer).split('\$');
}
