// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';

import '../../src/analysis/analyzer.dart';
import 'index_searcher.dart';

class BoolQuery {
  Op _op;
  BoolQuery([Op op]) {
    if (op == null) op = Op.or;
    _op = op;
  }
  List<Query> _queries = <Query>[];
  List<Query> _filters = <Query>[];
  BoolQuery addQuery(Query query) {
    _queries.add(query);
    return this;
  }

  BoolQuery addFilter(Query query) {
    _filters.add(query);
    return this;
  }

  static Analyzer defaultAnalyzer = new KeywordAnalyzer();
  BoolQuery append(String field, String text,
      {Analyzer analyzer, double boost: 1.00}) {
    final Analyzer ana = (analyzer == null) ? defaultAnalyzer : analyzer;
    for (String token in ana.getTokens(text)) {
      addQuery(new TermQuery(new Term(field, token, boost: boost)));
    }
    return this;
  }

  List<Query> get queries => _queries;
  List<Query> get filters => _filters;
  Op get op => _op;
}

enum Op { and, or }

/// The abstract base class for queries.
abstract class Query {
  Set<int> _matchedDocIds;
  Set<int> get matchedDocIds => _matchedDocIds;
  Future<Set<int>> match(IndexSearcher searcher) async {
    _matchedDocIds = await _matchDocIds(searcher);
    return _matchedDocIds;
  }

  Future<Set<int>> _matchDocIds(IndexSearcher searcher);
  double score();
}

class TermQuery extends Query {
  final Term _term;
  TermQuery(this._term);
  Term get term => _term;
  @override
  Future<Set<int>> _matchDocIds(IndexSearcher searcher) async {
    final Set<int> matched =
        (await searcher.reader.postings(_term.field, _term.text)).toSet();
    final Set<int> liveDocs = (await searcher.reader.liveDocs()).toSet();
    return IndexSearcher.unionDocIds(Op.and, matched, liveDocs);
  }

  @override
  double score() => (1.00 * _term.boost);
}

class Term {
  final String field;
  final String text;
  double boost;
  Term(this.field, this.text, {double boost: 1.00}) {
    this.boost = boost;
  }
}

class MatchAllDocsQuery extends Query {
  @override
  Future<Set<int>> _matchDocIds(IndexSearcher searcher) async {
    return (await searcher.reader.liveDocs()).toSet();
  }

  @override
  double score() => 1.00;
}

class MatchNoDocsQuery extends Query {
  @override
  Future<Set<int>> _matchDocIds(IndexSearcher searcher) async {
    return new Set<int>();
  }

  @override
  double score() => 0.00;
}

class RangeQuery extends Query {
  final String field;
  final double lower;
  final double upper;
  RangeQuery(this.field, this.lower, this.upper);
  @override
  Future<Set<int>> _matchDocIds(IndexSearcher searcher) async {
    Set<int> matched = new Set<int>();
    final Iterable<String> terms = await searcher.reader.terms(field);
    if (terms == null) return matched;
    bool isFirst = true;
    for (String term in terms) {
      final double value = double.parse(term);
      Set<int> docIds = new Set<int>();
      if (lower <= value && value <= upper) {
        docIds = (await searcher.reader.postings(field, term)).toSet();
      }
      if (isFirst) {
        matched = docIds;
        isFirst = false;
        continue;
      }
      matched = IndexSearcher.unionDocIds(Op.or, matched, docIds);
    }
    final Set<int> liveDocs = (await searcher.reader.liveDocs()).toSet();
    return IndexSearcher.unionDocIds(Op.and, matched, liveDocs);
  }

  @override
  double score() => 0.00;
}

class TermsQuery extends Query {
  final List<Term> _terms;
  TermsQuery(this._terms);
  @override
  Future<Set<int>> _matchDocIds(IndexSearcher searcher) async {
    Set<int> orIds = new Set<int>();
    for (Term term in _terms) {
      final TermQuery tq = new TermQuery(term);
      final Set<int> docIds = await tq.match(searcher);
      orIds = IndexSearcher.unionDocIds(Op.or, orIds, docIds);
    }
    return orIds;
  }

  @override
  double score() => 0.00;
}

class NotQuery extends Query {
  final List<Term> _prohibitTerms;
  NotQuery(this._prohibitTerms);
  @override
  Future<Set<int>> _matchDocIds(IndexSearcher searcher) async {
    Set<int> prohibitedIds = new Set<int>();
    final TermsQuery tsq = new TermsQuery(_prohibitTerms);
    prohibitedIds = await tsq.match(searcher);
    final Query q = new MatchAllDocsQuery();
    final Set<int> allIds = await q.match(searcher);
    return allIds.difference(prohibitedIds);
  }

  @override
  double score() => 0.00;
}
