// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';

import 'package:firebase/firebase.dart' as firebase;

import '../../../src/index/index_reader.dart';
import '../../../src/store/index_directory.dart';

class FirebaseIndexDirectory implements IndexDirectory {
  final FirebaseIndexStreams _indexStreams;
  final FirebaseDataStreams _dataStreams;
  FirebaseIndexDirectory(this._indexStreams, this._dataStreams);

  @override
  Future<Null> initOutput() async {}

  @override
  Future<Null> outputLiveDocs(Set<int> docIds) {
    throw new Exception('UnsupportedOperationException');
  }

  @override
  Future<Null> outputIndex(String name, Map<String, DocSet> index) {
    throw new Exception('UnsupportedOperationException');
  }

  @override
  Future<Null> outputData(String name, Map<int, DocData> data) {
    throw new Exception('UnsupportedOperationException');
  }

  @override
  void initInput() {}

  @override
  LiveDocsSource inputLiveDocs() {
    return new LiveDocsSource(new FirebaseLiveDocsStreams());
  }

  @override
  IndexSource inputIndex() {
    return new IndexSource(_indexStreams);
  }

  @override
  DataSource inputData() {
    return new DataSource(_dataStreams);
  }
}

class FirebaseLiveDocsStreams extends LiveDocsStreams {
  FirebaseLiveDocsStreams();
  @override
  Future<Null> readFully(DocSet to) async {
    // ignore: always_specify_types
    final ref = firebase.database().ref("liv");
    // ignore: always_specify_types
    final event = await ref.once("value");
    final Set<int> docIds = (event.snapshot.val() as dynamic)
        .split(",")
        .map((String s) => int.parse(s))
        .toSet();
    if (docIds.length > 0) {
      to.update(docIds);
    }
  }
}

class FirebaseIndexStreams extends IndexStreams {
  bool _readFully;
  bool loaded = false;

  @override
  Future<Null> readFully(Map<String, Map<String, DocSet>> to) async {
    _readFully = true;
    await FirebaseUtils._getFields("idx").then((List<String> fields) {
      for (String field in fields) {
        _getInvertedIndex("idx/" + field).then(
            (Map<String, DocSet> invertedIndex) => to[field] = invertedIndex);
      }
    }).then((_) {
      loaded = true;
    });
  }

  Future<Map<String, DocSet>> _getInvertedIndex(String path) async {
    final Map<String, DocSet> invertedIndex = <String, DocSet>{};
    // ignore: always_specify_types
    final query = firebase.database().ref(path).orderByKey();
    // ignore: always_specify_types
    final event = await query.once("value");
    // ignore: always_specify_types
    event.snapshot.forEach((e) {
      final String key = e.val()["term"];
      final String s = e.val()["postings"];
      final Set<int> values =
          s.split(',').map((String s) => int.parse(s)).toSet();
      invertedIndex[key] = new DocSet(values);
    });
    return invertedIndex;
  }

  @override
  Future<Null> readTerms(
      String field, Map<String, Map<String, DocSet>> to) async {
    if (!_readFully) await readFully(to);
  }

  @override
  Future<Null> readPostings(
      String field, String term, Map<String, Map<String, DocSet>> to) async {
    if (!_readFully) await readFully(to);
  }
}

class FirebaseDataStreams extends DataStreams {
  bool _readFully;
  bool loaded = false;

  @override
  Future<Null> readFully(Map<String, Map<int, DocData>> to) async {
    _readFully = true;
    await FirebaseUtils._getFields("data").then((List<String> fields) {
      for (String field in fields) {
        _getStoredData("data/" + field)
            .then((Map<int, DocData> storedData) => to[field] = storedData);
      }
    }).then((_) {
      loaded = true;
    });
  }

  Future<Map<int, DocData>> _getStoredData(String path) async {
    final Map<int, DocData> storedData = <int, DocData>{};
    // ignore: always_specify_types
    final query = firebase.database().ref(path).orderByKey();
    // ignore: always_specify_types
    final event = await query.once("value");
    // ignore: always_specify_types
    event.snapshot.forEach((e) {
      final int key = int.parse(e.key);
      final String s = e.val();
      final List<String> values = s.split('\$');
      storedData[key] = new DocData(values);
    });
    return storedData;
  }

  @override
  Future<Null> readFields(Map<String, Map<int, DocData>> to) async {
    if (!_readFully) await readFully(to);
  }

  @override
  Future<Null> readValues(
      String field, int docId, Map<String, Map<int, DocData>> to) async {
    if (!_readFully) await readFully(to);
  }
}

class FirebaseUtils {
  static Future<List<String>> _getFields(String extension) async {
    final List<String> fields = <String>[];
    // ignore: always_specify_types
    final query = firebase.database().ref(extension).orderByKey();
    // ignore: always_specify_types
    final event = await query.once("value");
    // ignore: always_specify_types
    event.snapshot.forEach((e) => fields.add(e.key));
    return fields;
  }
}
