// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';

import '../../src/document/document.dart';
import '../../src/index/index_reader.dart';
import '../../src/search/index_searcher.dart';
import '../../src/search/query.dart';
import '../../src/store/standard/composite_index_directory.dart';
import 'index_directory.dart';

/// An IndexWriter creates and maintains an index segment.
///
/// For multiple segments, use IndexHolder.
class IndexWriter {
  final IndexDirectory _directory;

  IndexWriter(this._directory);

  Future<Null> write(Iterable<Document> documents) async {
    final IndexTable table = buildIndexTable(documents);
    await writeTable(table);
  }

  Future<Null> writeTable(IndexTable table) async {
    await _directory.initOutput();
    await _storeLiveDocs(table.liveDocs);
    await _storeIndexes(table.indexes);
    await _storeData(table.data);
  }

  IndexTable buildIndexTable(Iterable<Document> documents) {
    final IndexTable table = new IndexTable();
    int docId = 0;
    for (Document doc in documents) {
      _buildIndexTable(docId, doc, table);
      docId++;
    }
    return table;
  }

  void _buildIndexTable(int docId, Document document, IndexTable table) {
    table.liveDocs.add(docId);
    for (Field field in document.fields) {
      final String name = field.name;
      final List<String> text = field.text;
      final Iterable<Iterable<String>> tokensIte =
          new Iterable<Iterable<String>>.generate(
              text.length, (int i) => field.analyzer.getTokens(text[i]));
      final Iterable<String> tokens =
          tokensIte.expand((Iterable<String> x) => x);
      if (!table.indexes.containsKey(name)) {
        table.indexes[name] = new InvertedIndexBuilder().append(tokens, docId);
      } else {
        table.indexes[name].append(tokens, docId);
      }
      if (field.stored) {
        if (!table.data.containsKey(name)) {
          table.data[name] = new StoredDataBuilder().append(docId, text);
        } else {
          table.data[name].append(docId, field.text);
        }
      }
    }
  }

  Future<Null> _storeLiveDocs(Set<int> liveDocs) async {
    await _directory.outputLiveDocs(liveDocs);
  }

  Future<Null> _storeIndexes(Map<String, InvertedIndexBuilder> indexes) async {
    for (String name in indexes.keys) {
      await _directory.outputIndex(name, indexes[name].build());
    }
  }

  Future<Null> _storeData(Map<String, StoredDataBuilder> data) async {
    for (String name in data.keys) {
      await _directory.outputData(name, data[name].build());
    }
  }

  Future<Null> delete(BoolQuery query) async {
    final IndexReader reader = await DirectoryReader.open(_directory);
    final IndexSearcher searcher = new IndexSearcher(reader);
    final TopDocs hits = await searcher.search(query, await reader.maxDoc(),
        scoreSort: ScoreSort.indexed());
    if (hits.totalHits > 0) {
      final Set<int> docIds =
          hits.scoreDocs.map((ScoreDoc scoreDoc) => scoreDoc.doc).toSet();
      final DocSet liveDocs = await reader.liveDocs();
      final Set<int> newLiveDocs = liveDocs.toSet().difference(docIds);
      await _storeLiveDocs(newLiveDocs);
    }
  }

  Future<Null> expungeDeletes() async {
    DocSet liveDocs = new DocSet(new Set<int>());
    liveDocs = await _directory.inputLiveDocs().read(liveDocs);

    Map<String, Map<String, DocSet>> indexes = <String, Map<String, DocSet>>{};
    indexes = await _directory.inputIndex().read(indexes);
    final Map<String, Map<String, DocSet>> newIndexes =
        <String, Map<String, DocSet>>{};
    expungeDeletesForIndexes(indexes, newIndexes, liveDocs);

    Map<String, Map<int, DocData>> data = <String, Map<int, DocData>>{};
    data = await _directory.inputData().read(data);
    final Map<String, Map<int, DocData>> newData =
        <String, Map<int, DocData>>{};
    expungeDeletesForData(data, newData, liveDocs);

    await _directory.initOutput();
    await _directory.outputLiveDocs(liveDocs.toSet());
    for (String field in newIndexes.keys) {
      await _directory.outputIndex(field, newIndexes[field]);
    }
    for (String field in newData.keys) {
      await _directory.outputData(field, newData[field]);
    }
  }

  static void expungeDeletesForIndexes(
      Map<String, Map<String, DocSet>> currentIndexes,
      Map<String, Map<String, DocSet>> newIndexes,
      DocSet liveDocs) {
    final Set<int> liveDocIds = liveDocs.toSet();
    for (String field in currentIndexes.keys) {
      final Map<String, DocSet> newIndex = <String, DocSet>{};
      final Map<String, DocSet> index = currentIndexes[field];
      for (String term in index.keys) {
        final Set<int> docIds = index[term].toSet();
        final Set<int> newLiveDocIds = docIds.intersection(liveDocIds);
        if (newLiveDocIds.length > 0) {
          final DocSet newDocSet = new DocSet(newLiveDocIds);
          newIndex[term] = newDocSet;
        }
      }
      newIndexes[field] = newIndex;
    }
  }

  static void expungeDeletesForData(Map<String, Map<int, DocData>> currentData,
      Map<String, Map<int, DocData>> newData, DocSet liveDocs) {
    for (String field in currentData.keys) {
      final Map<int, DocData> newDocs = <int, DocData>{};
      final Map<int, DocData> curDocs = currentData[field];
      for (int docId in curDocs.keys) {
        if (liveDocs.toSet().contains(docId)) {
          newDocs[docId] = curDocs[docId];
        }
      }
      newData[field] = newDocs;
    }
  }

  static Future<Null> copyFrom(
      CompositeIndexDirectory from, IndexDirectory to) async {
    await to.initOutput();
    DocSet liveDocs = new DocSet(new Set<int>());
    liveDocs = await from.inputLiveDocs().read(liveDocs);
    await to.outputLiveDocs(liveDocs.toSet());
    Map<String, Map<String, DocSet>> indexes = <String, Map<String, DocSet>>{};
    indexes = await from.inputIndex().read(indexes);
    for (String field in indexes.keys) {
      await to.outputIndex(field, indexes[field]);
    }
    Map<String, Map<int, DocData>> data = <String, Map<int, DocData>>{};
    data = await from.inputData().read(data);
    for (String field in data.keys) {
      await to.outputData(field, data[field]);
    }
  }
}

class IndexTable {
  Set<int> liveDocs = new Set<int>();
  Map<String, InvertedIndexBuilder> indexes = <String, InvertedIndexBuilder>{};
  Map<String, StoredDataBuilder> data = <String, StoredDataBuilder>{};
}

class InvertedIndexBuilder {
  Map<String, Set<int>> _invertedIndex = <String, Set<int>>{};

  InvertedIndexBuilder append(Iterable<String> tokens, int docId) {
    for (String token in tokens) {
      if (!_invertedIndex.containsKey(token)) {
        final Set<int> docIds = new Set<int>()..add(docId);
        _invertedIndex[token] = docIds;
      } else {
        final Set<int> docIds = _invertedIndex[token];
        if (!docIds.contains(docId)) {
          docIds.add(docId);
        }
      }
    }
    return this;
  }

  Map<String, DocSet> build() {
    final Map<String, DocSet> ret = <String, DocSet>{};
    final List<String> terms = _invertedIndex.keys.toList();
    terms.sort();
    for (String term in terms) {
      ret[term] = new DocSet(_invertedIndex[term]);
    }
    return ret;
  }
}

class StoredDataBuilder {
  Map<int, DocData> _storedData = <int, DocData>{};

  StoredDataBuilder append(int docId, List<String> text) {
    _storedData[docId] = new DocData(text);
    return this;
  }

  Map<int, DocData> build() {
    return _storedData;
  }
}
