// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';

import '../../../src/index/index_reader.dart';
import '../../../src/store/index_directory.dart';

/// Implements [IndexDirectory] composed of multiple [IndexDirectory] implementations.
class CompositeIndexDirectory implements IndexDirectory {
  final List<IndexDirectory> _directories;
  CompositeIndexDirectory(this._directories);
  @override
  Future<Null> initOutput() async {}
  @override
  Future<Null> outputLiveDocs(Set<int> docIds) {
    throw new Exception('UnsupportedOperationException');
  }

  @override
  Future<Null> outputIndex(String name, Map<String, DocSet> index) {
    throw new Exception('UnsupportedOperationException');
  }

  @override
  Future<Null> outputData(String name, Map<int, DocData> data) {
    throw new Exception('UnsupportedOperationException');
  }

  @override
  void initInput() {}
  @override
  LiveDocsSource inputLiveDocs() =>
      new LiveDocsSource(new CompositeLiveDocsStreams(_directories));
  @override
  IndexSource inputIndex() =>
      new IndexSource(new CompositeIndexStreams(_directories));
  @override
  DataSource inputData() =>
      new DataSource(new CompositeDataStreams(_directories));
}

class CompositeLiveDocsStreams extends LiveDocsStreams {
  final List<IndexDirectory> _directories;
  CompositeLiveDocsStreams(this._directories);
  @override
  Future<Null> readFully(DocSet to) async {
    final Set<int> ret = new Set<int>();
    int prevMaxDoc = 0;
    for (IndexDirectory directory in _directories) {
      Map<String, Map<String, DocSet>> indexes =
          <String, Map<String, DocSet>>{};
      indexes = await directory.inputIndex().read(indexes);
      final int maxDoc = indexes['id'].keys.length;
      DocSet docSet = new DocSet(new Set<int>());
      docSet = await directory.inputLiveDocs().read(docSet);
      final Set<int> docIds = docSet.toSet();
      merge(ret, docIds, prevMaxDoc);
      prevMaxDoc += maxDoc;
    }
    to.update(ret);
  }

  static void merge(
      Set<int> currentDocIds, Set<int> addDocIds, int prevMaxDoc) {
    currentDocIds
        .addAll(addDocIds.map((int docId) => docId + prevMaxDoc).toSet());
  }
}

class CompositeIndexStreams extends IndexStreams {
  final List<IndexDirectory> _directories;
  CompositeIndexStreams(this._directories);
  bool _readFully = false;
  @override
  Future<Null> readFully(Map<String, Map<String, DocSet>> to) async {
    _readFully = true;
    final Map<String, Map<String, DocSet>> ret =
        <String, Map<String, DocSet>>{};
    int prevMaxDoc = 0;
    for (IndexDirectory directory in _directories) {
      Map<String, Map<String, DocSet>> indexes =
          <String, Map<String, DocSet>>{};
      indexes = await directory.inputIndex().read(indexes);
      merge(ret, indexes, prevMaxDoc);
      final int maxDoc = indexes['id'].keys.length;
      prevMaxDoc += maxDoc;
    }
    for (String field in ret.keys) {
      to[field] = ret[field];
    }
  }

  static void merge(Map<String, Map<String, DocSet>> currentIndexes,
      Map<String, Map<String, DocSet>> addIndexes, int prevMaxDoc) {
    for (String field in addIndexes.keys) {
      if (currentIndexes[field] == null) {
        currentIndexes[field] = addIndexes[field];
        continue;
      }
      final Map<String, DocSet> index = addIndexes[field];
      for (String term in index.keys) {
        final Set<int> docIds = index[term].toSet();
        final Set<int> newDocIds =
            docIds.map((int id) => id + prevMaxDoc).toSet();
        if (currentIndexes[field][term] == null) {
          currentIndexes[field][term] = new DocSet(newDocIds);
        } else {
          final Set<int> cur = currentIndexes[field][term].toSet();
          cur.addAll(newDocIds);
          currentIndexes[field][term] = new DocSet(cur);
        }
      }
    }
  }

  @override
  Future<Null> readTerms(
      String field, Map<String, Map<String, DocSet>> to) async {
    if (!_readFully) await readFully(to);
  }

  @override
  Future<Null> readPostings(
      String field, String term, Map<String, Map<String, DocSet>> to) async {
    if (!_readFully) await readFully(to);
  }
}

class CompositeDataStreams extends DataStreams {
  final List<IndexDirectory> _directories;
  CompositeDataStreams(this._directories);
  bool _readFully = false;
  @override
  Future<Null> readFully(Map<String, Map<int, DocData>> to) async {
    _readFully = true;
    final Map<String, Map<int, DocData>> ret = <String, Map<int, DocData>>{};
    int prevMaxDoc = 0;
    for (IndexDirectory directory in _directories) {
      Map<String, Map<int, DocData>> data = <String, Map<int, DocData>>{};
      data = await directory.inputData().read(data);
      merge(ret, data, prevMaxDoc);
      final int maxDoc = data['id'].keys.length;
      prevMaxDoc += maxDoc;
    }
    for (String field in ret.keys) {
      to[field] = ret[field];
    }
  }

  static void merge(Map<String, Map<int, DocData>> currentData,
      Map<String, Map<int, DocData>> addData, int prevMaxDoc) {
    for (String field in addData.keys) {
      if (currentData[field] == null) {
        currentData[field] = addData[field];
        continue;
      }
      final Map<int, DocData> docs = addData[field];
      for (int docId in docs.keys) {
        final int newDocIds = docId + prevMaxDoc;
        currentData[field][newDocIds] = docs[docId];
      }
    }
  }

  @override
  Future<Null> readFields(Map<String, Map<int, DocData>> to) async {
    if (!_readFully) await readFully(to);
  }

  @override
  Future<Null> readValues(
      String field, int docId, Map<String, Map<int, DocData>> to) async {
    if (!_readFully) await readFully(to);
  }
}
