// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';
import 'dart:convert';
import 'dart:io';

import '../../../src/index/index_reader.dart';
import '../../../src/store/index_directory.dart';

/// A file system based [IndexDirectory] implementation.
class FSIndexDirectory implements IndexDirectory {
  final Directory _dir;
  FSIndexDirectory(this._dir);
  Directory get dir => _dir;

  @override
  Future<Null> initOutput() async {
    if (_dir.existsSync()) _dir.deleteSync(recursive: true);
    if (!_dir.existsSync()) _dir.createSync(recursive: true);
  }

  @override
  Future<Null> outputLiveDocs(Set<int> docIds) async {
    final File liveDocsFile = new File('${_dir.path}/.liv');
    if (liveDocsFile.existsSync()) liveDocsFile.deleteSync(); //mutable
    if (!liveDocsFile.existsSync()) liveDocsFile.createSync();
    liveDocsFile.writeAsStringSync(docIds.join(','),
        mode: FileMode.APPEND, encoding: UTF8);
  }

  @override
  Future<Null> outputIndex(String name, Map<String, DocSet> index) async {
    final File indexFile = new File('${_dir.path}/$name.idx');
    if (!indexFile.existsSync()) indexFile.createSync();
    for (String key in index.keys) {
      final String value = index[key].toSet().join(',');
      final String line = '$key=>$value';
      indexFile.writeAsStringSync(line + '\n',
          mode: FileMode.APPEND, encoding: UTF8);
    }
  }

  @override
  Future<Null> outputData(String name, Map<int, DocData> data) async {
    final File storeFile = new File('${_dir.path}/$name.data');
    if (!storeFile.existsSync()) storeFile.createSync();
    for (int key in data.keys) {
      final String value = data[key].toList().join('\$');
      final String line = '${key.toString()}=>$value';
      storeFile.writeAsStringSync(line + '\n',
          mode: FileMode.APPEND, encoding: UTF8);
    }
  }

  @override
  void initInput() {}
  @override
  LiveDocsSource inputLiveDocs() =>
      new LiveDocsSource(new FSLiveDocsStreams(_dir));
  @override
  IndexSource inputIndex() => new IndexSource(new FSIndexStreams(_dir));
  @override
  DataSource inputData() => new DataSource(new FSDataStreams(_dir));
}

class FSLiveDocsStreams extends LiveDocsStreams {
  final Directory _dir;
  FSLiveDocsStreams(this._dir);
  @override
  Future<Null> readFully(DocSet to) async {
    for (String path in FSIndexUtils.getFileList(_dir, '.liv')) {
      final File indexFile = new File(path);
      final List<String> lines = indexFile.readAsLinesSync();
      if (lines.length == 0) break;
      final String s = lines[0];
      final Set<int> docIds =
          s.split(',').map((String s) => int.parse(s)).toSet();
      to.update(docIds);
      break;
    }
  }
}

class FSIndexStreams extends IndexStreams {
  final Directory _dir;
  FSIndexStreams(this._dir);
  bool _readFully = false;
  @override
  Future<Null> readFully(Map<String, Map<String, DocSet>> to) async {
    for (String path in FSIndexUtils.getFileList(_dir, '.idx')) {
      final String name = FSIndexUtils.getName(path, '.idx');
      final File indexFile = new File(path);
      final List<String> lines = indexFile.readAsLinesSync();
      final Map<String, DocSet> invertedIndex = <String, DocSet>{};
      for (String line in lines) {
        final List<String> kv = line.split('=>');
        final String key = kv[0];
        final Set<int> values =
            kv[1].split(',').map((String s) => int.parse(s)).toSet();
        invertedIndex[key] = new DocSet(values);
      }
      to[name] = invertedIndex;
    }
    _readFully = true;
  }

  @override
  Future<Null> readTerms(
      String field, Map<String, Map<String, DocSet>> to) async {
    if (!_readFully) await readFully(to);
  }

  @override
  Future<Null> readPostings(
      String field, String term, Map<String, Map<String, DocSet>> to) async {
    if (!_readFully) await readFully(to);
  }
}

class FSDataStreams extends DataStreams {
  final Directory _dir;
  FSDataStreams(this._dir);
  bool _readFully = false;
  @override
  Future<Null> readFully(Map<String, Map<int, DocData>> to) async {
    for (String path in FSIndexUtils.getFileList(_dir, '.data')) {
      final String name = FSIndexUtils.getName(path, '.data');
      final File dataFile = new File(path);
      final String str = dataFile.readAsStringSync();
      final List<String> lines = str.split('\n');
      final Map<int, DocData> storedData = <int, DocData>{};
      for (String line in lines) {
        final List<String> kv = line.split('=>');
        if (kv.length == 2) {
          final int key = int.parse(kv[0]);
          final List<String> values = kv[1].split('\$');
          storedData[key] = new DocData(values);
        }
      }
      to[name] = storedData;
    }
    _readFully = true;
  }

  @override
  Future<Null> readFields(Map<String, Map<int, DocData>> to) async {
    if (!_readFully) await readFully(to);
  }

  @override
  Future<Null> readValues(
      String field, int docId, Map<String, Map<int, DocData>> to) async {
    if (!_readFully) await readFully(to);
  }
}

class FSIndexUtils {
  static Iterable<String> getFileList(Directory dir, String extension) => dir
      .listSync(recursive: false, followLinks: false)
      .map((FileSystemEntity e) => e.path)
      .where((String path) => path.endsWith(extension));

  static String getName(String path, String extension) =>
      path.split('/').last.replaceAll(extension, '');
}

class FSIndexJsonConverter {
  final Directory _dir;
  FSIndexDirectory _fsDir;

  FSIndexJsonConverter(this._dir) {
    _fsDir = new FSIndexDirectory(_dir);
  }

  Future<Null> run() async {
    final Map<String, Object> ret = <String, Object>{};

    DocSet docSet = new DocSet(new Set<int>());
    docSet = await _fsDir.inputLiveDocs().read(docSet);
    final Set<int> docIds = docSet.toSet();
    final String liveDocs = docIds.join(",");

    Map<String, Map<String, DocSet>> inputIndex =
        <String, Map<String, DocSet>>{};
    inputIndex = await _fsDir.inputIndex().read(inputIndex);
    final Map<String, Map<String, Map<String, String>>> idx =
        <String, Map<String, Map<String, String>>>{};
    for (String fieldName in inputIndex.keys) {
      final Map<String, DocSet> fieldValue = inputIndex[fieldName];
      final Map<String, Map<String, String>> terms =
          <String, Map<String, String>>{};
      int i = 0;
      for (String term in fieldValue.keys) {
        final Map<String, String> termInfo = <String, String>{};
        final Set<int> postings = fieldValue[term].toSet();
        termInfo["postings"] = postings.join(",");
        termInfo["term"] = term;
        terms[i.toString()] = termInfo;
        i++;
      }
      idx[fieldName] = terms;
    }

    Map<String, Map<int, DocData>> inputData = <String, Map<int, DocData>>{};
    inputData = await _fsDir.inputData().read(inputData);
    final Map<String, Map<String, String>> data =
        <String, Map<String, String>>{};
    for (String fieldName in inputData.keys) {
      final Map<int, DocData> fieldValue = inputData[fieldName];
      final Map<String, String> fieldData = <String, String>{};
      for (int docId in fieldValue.keys) {
        final List<String> data = fieldValue[docId].toList();
        fieldData[docId.toString()] = data.join("\$");
      }
      data[fieldName] = fieldData;
    }

    ret["idx"] = idx;
    ret["data"] = data;
    ret["liv"] = liveDocs;

    _outJson(ret, "fsindex.json");
  }

  void _outJson(Map<String, Object> obj, String outFilePath) {
    final File outFile = new File(outFilePath);
    if (outFile.existsSync()) outFile.deleteSync();
    if (!outFile.existsSync()) outFile.createSync();
    outFile.writeAsStringSync(JSON.encode(obj),
        mode: FileMode.APPEND, encoding: UTF8);
  }
}
