// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';

import '../../../src/index/index_reader.dart';
import '../../../src/store/index_directory.dart';

/// A memory-resident [IndexDirectory] implementation.
class RAMIndexDirectory implements IndexDirectory {
  DocSet _liveDocs = new DocSet(new Set<int>());
  Map<String, Map<String, DocSet>> _ramIndex = <String, Map<String, DocSet>>{};
  Map<String, Map<int, DocData>> _ramData = <String, Map<int, DocData>>{};

  @override
  Future<Null> initOutput() async {}

  @override
  Future<Null> outputLiveDocs(Set<int> docIds) async {
    _liveDocs = new DocSet(docIds);
  }

  @override
  Future<Null> outputIndex(String name, Map<String, DocSet> index) async {
    _ramIndex[name] = index;
  }

  @override
  Future<Null> outputData(String name, Map<int, DocData> data) async {
    _ramData[name] = data;
  }

  @override
  void initInput() {}
  @override
  LiveDocsSource inputLiveDocs() =>
      new LiveDocsSource(new RAMLiveDocsStreams(_liveDocs));
  @override
  IndexSource inputIndex() => new IndexSource(new RAMIndexStreams(_ramIndex));
  @override
  DataSource inputData() => new DataSource(new RAMDataStreams(_ramData));
}

class RAMLiveDocsStreams extends LiveDocsStreams {
  final DocSet _liveDocs;
  RAMLiveDocsStreams(this._liveDocs);
  @override
  Future<Null> readFully(DocSet to) async {
    to.update(_liveDocs.toSet());
  }
}

class RAMIndexStreams extends IndexStreams {
  final Map<String, Map<String, DocSet>> _ramIndex;
  RAMIndexStreams(this._ramIndex);
  bool _readFully = false;
  @override
  Future<Null> readFully(Map<String, Map<String, DocSet>> to) async {
    for (String field in _ramIndex.keys) {
      to[field] = _ramIndex[field];
    }
    _readFully = true;
  }

  @override
  Future<Null> readTerms(
      String field, Map<String, Map<String, DocSet>> to) async {
    if (!_readFully) await readFully(to);
  }

  @override
  Future<Null> readPostings(
      String field, String term, Map<String, Map<String, DocSet>> to) async {
    if (!_readFully) await readFully(to);
  }
}

class RAMDataStreams extends DataStreams {
  final Map<String, Map<int, DocData>> _ramData;
  RAMDataStreams(this._ramData);
  bool _readFully = false;
  @override
  Future<Null> readFully(Map<String, Map<int, DocData>> to) async {
    for (String field in _ramData.keys) {
      to[field] = _ramData[field];
    }
    _readFully = true;
  }

  @override
  Future<Null> readFields(Map<String, Map<int, DocData>> to) async {
    if (!_readFully) await readFully(to);
  }

  @override
  Future<Null> readValues(
      String field, int docId, Map<String, Map<int, DocData>> to) async {
    if (!_readFully) await readFully(to);
  }
}
